#!/bin/python

import sys
import os
import subprocess
import json

import sdk
import user

BIN_RSYNC="/bin/rsync"
BIN_TAR="/bin/tar"
BIN_RM="/bin/rm"
BIN_CP="/bin/cp"
BIN_SED="/bin/sed"
BIN_MV="/bin/mv"
BIN_SQLITE3="/bin/sqlite3"
BIN_SYNO_SKV="/usr/syno/bin/synosetkeyvalue"
BIN_SYNO_GKV="/usr/syno/bin/synogetkeyvalue"
BIN_JSON_CLEANER="/usr/syno/etc/synomigration/scripts/python/space/json_cleaner.py"

SZF_RSYNC_FILTER_DIR="/usr/syno/etc/synomigration/scripts/python/space/filter/"
CFG_SYNO_MIGRATION="/usr/syno/etc/synomigration/info.cfg"

SZF_ARCHIVE_DIR="/syno_migrate/"
SZD_OLDPATCH_DIR=SZF_ARCHIVE_DIR+".old_patch_info/"
SZF_ARCHIVE_PATH=SZF_ARCHIVE_DIR+".syno_migration_archive.tar"
SZF_OLDPATCH_PATH=SZF_ARCHIVE_DIR+".syno_old_patch.tar"

def CreateFolderIfNotExisted(folder_path):
    if os.path.isdir(folder_path):
        return True

    if os.path.exists(folder_path):
        return False

    os.mkdir(folder_path)

    return True


def RsyncExecutor(cmd_arg):
    try:
        retcode = subprocess.call(cmd_arg, stdout=subprocess.PIPE)

        return retcode
    except OSError as e:
        log("{0} execution failed: {1}".format(cmd_arg[0], e))
        return -1


def ArchiveExecutor(archive_folder):

    if not archive_folder:
        return False

    filter_file = "{0}{1}.filter".format(SZF_RSYNC_FILTER_DIR, archive_folder)
    if not os.path.exists(filter_file):
        return False


    sdk.log("starting archive {0}".format(archive_folder))

    cmd_arg = [
        BIN_RSYNC,
        "-a",
        "--exclude-from={0}".format(filter_file),
        "/" + archive_folder + "/",
        "{0}{1}".format(SZF_ARCHIVE_DIR, archive_folder)
    ]
    retcode = RsyncExecutor(cmd_arg)

    # rsync error code:
    #  - 23: partial - some file attribute is not copied
    #  - 24: vanished - file is removed between file list and copy
    if retcode != 0 and retcode != 24:
        sdk.log("{0} was returned: {1}".format(cmd_arg[0], retcode))
        return False

    if retcode == 24:
        sdk.log("rsync is returned 24 (vanished)")

    return True


def ClearUDC():
    '''
    clear following settings in /etc/synoinfo.conf
     - enable_data_collect="yes"
     - udc_check_state="6.2.2"
    '''
    if not sdk.CmdExecutor([BIN_SYNO_SKV, SZF_ARCHIVE_DIR+"etc/synoinfo.conf", "enable_data_collect", ""]):
        sdk.log("fail to clear enable_data_collect in synoinfo.conf")

    if not sdk.CmdExecutor([BIN_SYNO_SKV, SZF_ARCHIVE_DIR+"etc/synoinfo.conf", "udc_check_state", ""]):
        sdk.log("fail to clear udc_check_state in synoinfo.conf")


def ClearPkgTerm():
    '''
    clear package term by executing for every user
    sed -i 's/"agreed_term_version.*,//g' /usr/syno/etc/preference/<username>/usersettings
    '''
    preference_path = SZF_ARCHIVE_DIR+"/usr/syno/etc/preference/"
    for root, dirs, files in os.walk(preference_path):
        for dir_ in dirs:
            user_preference_path = preference_path+dir_+"/usersettings"
            if os.path.exists(user_preference_path):
                if not sdk.CmdExecutor([BIN_SED, "-i", 's/"agreed_term_version.*,//g', user_preference_path]):
                    sdk.log("fail to clear package term for {}".format(user))


def FilterSettings():
    '''
    some settings should not be migrated due to privacy:
     - UDC
     - Package Term
     - License
    '''
    ClearUDC()
    ClearPkgTerm()


def DisableUpstartService(service_name):
    service_override = SZF_ARCHIVE_DIR+"etc/init/{}.override".format(service_name)
    if not os.path.exists(service_override):
        with open(service_override, 'w') as f:
            f.write("manual")


def SetAppPrivilegeDefaultRule():
    # check default rule existence
    select_cmd = "SELECT * FROM AppPrivRule WHERE App == 'SYNO.Finder.Application'"
    sql_cmd_arg = [BIN_SQLITE3, SZF_ARCHIVE_DIR+"etc/synoappprivilege.db", select_cmd]
    finder_rule = subprocess.Popen(sql_cmd_arg, stdout=subprocess.PIPE)
    if finder_rule.stdout.readline():
        return True

    # insert finder default rule
    insert_cmd = "INSERT INTO AppPrivRule VALUES ({}, {}, {}, {}, {}, '', '')".format(
        1,
        101,
        "'SYNO.Finder.Application'",
        "'0.0.0.0'",
        "'0000:0000:0000:0000:0000:FFFF:0000:0000'")
    sql_cmd_arg = [BIN_SQLITE3, SZF_ARCHIVE_DIR+"etc/synoappprivilege.db", insert_cmd]
    if not sdk.CmdExecutor(sql_cmd_arg):
        sdk.log("fail to insert SynoFinder default rules")
        return False

    return True


def LogoutSynologyAccount():
    if os.path.exists(SZF_ARCHIVE_DIR+"etc/myds.conf"):
        if not sdk.CmdExecutor([BIN_SYNO_SKV, SZF_ARCHIVE_DIR+"etc/myds.conf", "mydsIDSerial", ""]):
            sdk.log("fail to clear mydsIDSerial in myds.conf")


def OverrideServiceConfig(cfg_file, cfg):
    service_override_path = SZF_ARCHIVE_DIR + "usr/syno/etc/synoservice.override/"

    # add config if override config is not existing
    if not os.path.isfile(service_override_path + cfg_file):
        with open(service_override_path + cfg_file, 'w') as file:
            try:
                json.dump(cfg, file)
            except:
                sdk.log("fail to dump json file [{}]".format(
                    service_override_path + cfg_file))


def FixInvalidJSON(cfg_file):
    if not os.path.exists(cfg_file):
        return True

    if not sdk.CmdExecutor([BIN_JSON_CLEANER, cfg_file]):
        sdk.log("fail to clean invalid json config {}".format(cfg_file))
        return False

    return True


def OverrideAllServiceConfig():
    service_path = SZF_ARCHIVE_DIR + "usr/syno/etc/synoservice.d/"

    for cfg_file in os.listdir(service_path):
        if not os.path.isfile(service_path + cfg_file):
            continue

        with open(service_path + cfg_file, 'r') as file:
            try:
                config = json.load(file)
            except:
                config = {}

        # retry for invalid json (remove trailing comma)
        if not config:
            FixInvalidJSON(service_path + cfg_file)
            with open(service_path + cfg_file, 'r') as file:
                try:
                    config = json.load(file)
                except:
                    config = {}
                    sdk.log("fail to load file as json format [{}]".format(service_path + cfg_file))


        if "" != config.get("auto_start", ""):
            OverrideServiceConfig(cfg_file, {"auto_start": config["auto_start"]})


def DisableDHCPServer():
    # disable DHCP server by setting /etc/dhcpd/dhcpd-$IF.info to no

    dhcp_config_dir = SZF_ARCHIVE_DIR + "etc/dhcpd"

    if not os.path.exists(dhcp_config_dir):
        return True

    if not os.path.isdir(dhcp_config_dir):
        return False

    for cfg_file in os.listdir(dhcp_config_dir):
        if not cfg_file.startswith("dhcpd-") or not cfg_file.endswith(".info"):
            continue

        # skip subnet setting
        if "subnet" in cfg_file:
            continue

        full_path = dhcp_config_dir + "/" + cfg_file
        if not sdk.CmdExecutor([BIN_SYNO_SKV, full_path, "enable", "no"]):
            sdk.log("fail to disable DHCP server of {}".format(full_path))

    return True


def DisableFireWallByConfig(config_path):

    # load firewall config into json
    with open(config_path, 'r') as fd_cfg:
        try:
            json_cfg = json.load(fd_cfg)
        except Exception:
            sdk.log("fail to load file as json format [{}]".format(config_path))
            return False

    # disable firewall
    json_cfg["status"] = False

    # write json to temp file
    tmp_path = config_path + ".tmp"
    with open(tmp_path, 'w') as fd_tmp:
        try:
            json.dump(json_cfg, fd_tmp, sort_keys=True, indent=4)
        except Exception:
            sdk.log("fail to dump json file [{}]".format(tmp_path))
            return False

    # rename to original config
    os.unlink(config_path)
    os.rename(tmp_path, config_path)

    return True


def DisableFireWall():
    # disable firewall by setting /usr/syno/etc/firewall.d/firewall_settings.json

    firewall_config_dir = SZF_ARCHIVE_DIR + "usr/syno/etc/firewall.d"

    if not os.path.exists(firewall_config_dir):
        return True

    if not os.path.isdir(firewall_config_dir):
        return False

    # disable firewall
    full_path = firewall_config_dir + "/firewall_settings.json"
    if not DisableFireWallByConfig(full_path):
        sdk.log("fail to disable firewall of {}".format(full_path))
        return False

    return True


def WebapiSRPlanEnum():
    '''
    synowebapi --exec api=SYNO.DR.Plan method=list version=1
        additional='["sync_policy"]'
    '''

    plan_list = []

    kwargs = {
        'additional': ["sync_policy"]
    }
    j_data = sdk.execWebAPI('SYNO.DR.Plan', 1, 'list', **kwargs)

    if not j_data or not j_data["success"]:
        return plan_list

    if not j_data["data"] or not j_data["data"]["plans"] or not isinstance(j_data["data"]["plans"], list):
        return plan_list

    # compose plan list contains scheduled to remote server
    for plan in j_data["data"]["plans"]:
        if plan["plan_id"] == plan["main_site"] or plan["plan_id"] == plan["dr_site"]:
            continue

        if not plan["additional"]["sync_policy"]:
            continue

        if plan["additional"]["sync_policy"]["enabled"]:
            plan["additional"]["sync_policy"]["enabled"] = False
            plan_list.append(plan)

    return plan_list


def WebapiSRScheduleDisable(plan):
    '''
    synowebapi --exec api=SYNO.DR.Plan.MainSite method=edit version=1
        plan_id=7a2caff4-c358-4860-8062-4908781ddb4d sync_policy='{}'
    '''

    kwargs = {
        'plan_id': plan["plan_id"],
        'sync_policy': plan["additional"]["sync_policy"]
    }
    j_data = sdk.execWebAPI('SYNO.DR.Plan', 1, 'edit', **kwargs)

    if not j_data or not j_data["success"]:
        return False

    return True


def DisableSRSchedule():
    ret = True

    for plan in WebapiSRPlanEnum():
        if not WebapiSRScheduleDisable(plan):
            sdk.log("fail to disable SR plan {}".format(plan["plan_id"]))
            ret = False

    return ret


def GetRemoteModel():
    return GetRemoteConfig("remote_model")


def GetRemoteConfig(config_name):
    # read config
    mgr_cfg = open(CFG_SYNO_MIGRATION, mode='r')
    if not mgr_cfg:
        sdk.log("fail to open migration config file: {}".format(CFG_SYNO_MIGRATION))
        return ""

    # load into json
    mgr_json = json.load(mgr_cfg)
    if not mgr_cfg:
        sdk.log("fail to load migration json: {}".format(CFG_SYNO_MIGRATION))
        return ""

    return mgr_json[config_name]


def DisablePowerSchedule():
    # check local synoinfo capability (support_power_schedule)
    if not GetRemoteModel() in ["synology_broadwellntbap_sa3200d", "synology_bromolow_rc18015xs+"]:
        return True

    # remove power schecdule config
    if os.path.exists(SZF_ARCHIVE_DIR+"etc/power_sched.conf"):
        os.unlink(SZF_ARCHIVE_DIR+"etc/power_sched.conf")

    # remove power schedule if target does not support
    if not sdk.CmdExecutor([BIN_SED, "-i", "/synopoweroff/d", SZF_ARCHIVE_DIR+"etc/crontab"]):
        sdk.log("fail to remove power off task")
        return False

    return True


def HandleXASynoinfo():
    '''
    DS -> HA: enable runha config
    XA -> DS: disable runha config
    '''

    # adjust runha run key according to target model
    runha = 'yes' if "broadwellntbap" in GetRemoteModel() else 'no'
    if not sdk.CmdExecutor([BIN_SYNO_SKV, SZF_ARCHIVE_DIR+"etc/synoinfo.conf", "runha", runha]):
        sdk.log("fail to set runha in synoinfo.conf")
        return False

    return True


def DisableWOL():
    '''
    /etc/synoinfo.conf:eth3_wol_options="d"
    /etc/synoinfo.conf:eth1_wol_options="g" -> "d"
    /etc/synoinfo.conf:eth0_wol_options="d"
    /etc/synoinfo.conf:eth2_wol_options="d"
    /etc/synoinfo.conf:wol_enabled_options="g" -> "d"
    '''

    key_list = ["wol_enabled_options"]
    for i in range(4):
        key_list.append("eth{}_wol_options".format(i))

    for key in key_list:
        setting = subprocess.Popen([BIN_SYNO_GKV, SZF_ARCHIVE_DIR+"etc/synoinfo.conf", key], stdout=subprocess.PIPE)

        if 'g' == setting.stdout.readline().rstrip():
            if not sdk.CmdExecutor([BIN_SYNO_SKV, SZF_ARCHIVE_DIR+"etc/synoinfo.conf", key, "d"]):
                sdk.log("fail to disable WOL")
                return False

    return True


def EnableNtpServer():
    '''
    /usr/syno/etc/synoservice.override/ntpd-server.cfg ("auto_start":"yes")
    '''

    # read service config
    json_cfg = {}
    cfg = SZF_ARCHIVE_DIR+"usr/syno/etc/synoservice.override/ntpd-server.cfg"
    if os.path.exists(cfg):
        with open(cfg, 'r') as fd_cfg:
            try:
                json_cfg = json.load(fd_cfg)
            except Exception:
                sdk.log("fail to load file as json format [{}]".format(cfg))
                return False

    # enable ntp server
    json_cfg["auto_start"] = "yes"

    # update service config
    tmp_path = cfg + ".tmp"
    with open(tmp_path, 'w') as fd_tmp:
        try:
            json.dump(json_cfg, fd_tmp, sort_keys=True, indent=4)
        except Exception:
            sdk.log("fail to dump json file [{}]".format(tmp_path))
            return False

    # rename to original config
    if os.path.exists(cfg):
        os.unlink(cfg)
    os.rename(tmp_path, cfg)

    return True


def HandleXAConfig():
    #remove /etc/init/synopkgctl-launcher.override if remote is not XA
    if GetRemoteModel() != "synology_broadwellntbap_sa3200d":
        cfg_list = [
            "etc/init/synopkgctl-launcher.override",
            "usr/share/init/synopkgctl-launcher.override"
        ]
        for cfg in cfg_list:
            full_path = SZF_ARCHIVE_DIR+cfg
            if os.path.exists(full_path):
                os.unlink(full_path)
    else:
        # handle XA default settings
        DisableWOL()
        EnableNtpServer()

    return True


def RemovePackageUser():
    '''
    remove package user (MigrationAssistant) in /etc/passwd, shadow
    '''

    if not sdk.CmdExecutor([BIN_SED, "-i", "/MigrationAssistant/d", SZF_ARCHIVE_DIR+"etc/passwd"]):
        sdk.log("fail to remove package user")
        return False

    if not sdk.CmdExecutor([BIN_SED, "-i", "/MigrationAssistant/d", SZF_ARCHIVE_DIR+"etc/shadow"]):
        sdk.log("fail to remove package user")
        return False

    return True


def GetDHCPServerIfList():
    '''
    list all DHCP server interface by glob /etc/dhcpd/dhcpd-*.info
    '''
    import glob

    prefix = SZF_ARCHIVE_DIR + 'etc/dhcpd/dhcpd-'
    posfix = ".info"
    config_list = glob.glob(prefix + '*' + posfix)

    if_list = []
    for c in config_list:
        if "subnet" in c:
            continue

        if_list.append(c[len(prefix):-len(posfix)])

    return if_list


def renameDHCPServerConfig(if_org, if_new):
    '''
    rename following files according to netsdk (libsynonetsdk/lib/c/DHCP/net_dhcp_server_change_interface.c)
        dhcpd-$IF.info
        dhcpd-$IF-subnet[n].conf   # need replace tag
        dhcpd-$IF-subnet[n].info
        dhcpd-$IF-static.conf
        wpad-$IF.conf              # need replace tag
        wpad-$IF.info
    '''

    def pattern_rename(pattern, if_org, if_new, replace_tag=False):
        config_org = pattern.format(if_org)
        config_new = pattern.format(if_new)
        if os.path.exists(config_org):
            os.rename(config_org, config_new)

        if replace_tag:
            if not sdk.CmdExecutor([BIN_SED, "-i", "s/{}/{}/g".format(if_org, if_new), config_new]):
                sdk.log("fail to replace tag: {}".format(config_new))


    # global
    pattern_rename(SZF_ARCHIVE_DIR + "etc/dhcpd/dhcpd-{}.info", if_org, if_new)

    # tag
    for i in range(100):
        tag_config = SZF_ARCHIVE_DIR + "etc/dhcpd/dhcpd-{}-subnet{}.conf".format(if_org, i)
        if not os.path.exists(tag_config):
            break;
        pattern_rename(SZF_ARCHIVE_DIR + "etc/dhcpd/dhcpd-{}" + "-subnet{}.conf".format(i), if_org, if_new, True)
        pattern_rename(SZF_ARCHIVE_DIR + "etc/dhcpd/dhcpd-{}" + "-subnet{}.info".format(i), if_org, if_new)

    # static
    pattern_rename(SZF_ARCHIVE_DIR + "etc/dhcpd/dhcpd-{}-static.conf", if_org, if_new)

    # wpad
    pattern_rename(SZF_ARCHIVE_DIR + "etc/dhcpd/wpad-{}.conf", if_org, if_new, True)
    pattern_rename(SZF_ARCHIVE_DIR + "etc/dhcpd/wpad-{}.info", if_org, if_new)

    return True


def AdjustDHCPServerConfig():
    '''
    DHCP server config contains interface name, which might be different after enabling OVS
    '''

    # check ovs setting on both side
    remote_ovs_enabled = GetRemoteConfig("remote_ovs")
    if "" == remote_ovs_enabled:
        return False

    local_ovs_enabled = os.path.exists("/usr/syno/etc/synoovs/ovs_reg.conf")
    if local_ovs_enabled == (1 == remote_ovs_enabled):
        return True

    # get interface list
    if_list = GetDHCPServerIfList()
    if not if_list:
        return True

    # change interface
    for org_if in if_list:

        if (1 == remote_ovs_enabled) == ("ovs" in org_if):
            continue

        if remote_ovs_enabled:
            new_if = "ovs_" + org_if
        else:
            new_if = org_if[len("ovs_"):]

        print "{} to {}".format(org_if, new_if)
        if not renameDHCPServerConfig(org_if, new_if):
            sdk.log("fail to rename DHCP server config: {}".format(org_if))

    return True


def FlushSharingDatabase():
    sharing_database = "/usr/syno/etc/private/session/sharing.db"
    if not sdk.CmdExecutor([BIN_SQLITE3, sharing_database, "PRAGMA wal_checkpoint"]):
        sdk.log("fail to flush sharing database: {}".format(config_new))


def SystemConfigArchive():

    # clear previous result
    sdk.CmdExecutor([BIN_RM, "-rf", SZF_ARCHIVE_DIR])

    # check archive folder existency
    if not CreateFolderIfNotExisted(SZF_ARCHIVE_DIR):
        sdk.log("fail to create archive directory: ({})".format(SZF_ARCHIVE_DIR))
        return False

    DisableSRSchedule()                 # disable Snapshot Replication schedule

    # archive config by rsync
    archive_list = ["etc", "usr", "var"]
    for d in archive_list:
        if not ArchiveExecutor(d):
            sdk.log("fail to archive: ({})".format(d))
            return False

    FilterSettings()                    # filter settings (UDC, package term)
    LogoutSynologyAccount()             # logout Synology Account
    DisableUpstartService("synorelayd") # disable quick connect due to synology account
    SetAppPrivilegeDefaultRule()        # set built-in package application privilege
    OverrideAllServiceConfig()          # add override config for services
    DisableDHCPServer()                 # disable DHCP server
    AdjustDHCPServerConfig()            # adjust DHCP server config according to interface name
    DisableFireWall()                   # disable firewall rules
    DisablePowerSchedule()              # disable power schedule if not supported
    HandleXASynoinfo()                  # keep runha key in synoinfo.conf
    HandleXAConfig()                    # remove /etc/init/synopkgctl-launcher.override if any
    RemovePackageUser()                 # remove package user
    FlushSharingDatabase()              # flush sharing database journal

    # tar needed files
    archive_dest=[]
    for d in archive_list:
        archive_dest.append("{0}{1}".format(SZF_ARCHIVE_DIR, d))

    tar_cmd_arg = [BIN_TAR, "-cf", SZF_ARCHIVE_PATH, "-C", SZF_ARCHIVE_DIR]
    tar_cmd_arg.extend(archive_list)
    if not sdk.CmdExecutor(tar_cmd_arg):
        sdk.log("fail to tar system config")
        return False

    return True


def SystemInfoArchive():

    # check archive folder existency
    if not CreateFolderIfNotExisted(SZD_OLDPATCH_DIR):
        sdk.log("fail to create directory: ({})".format(SZD_OLDPATCH_DIR))
        return False

    # cp /etc.defaults/synoinfo.conf /.old_patch_info/
    # cp /etc.defaults/VERSION /.old_patch_info/
    # cp /.system_info /.old_patch_info/
    if not sdk.CmdExecutor([BIN_CP, "-a", "/etc.defaults/synoinfo.conf", SZD_OLDPATCH_DIR]):
        sdk.log("fail to copy synoinfo.conf")
        return False
    if not sdk.CmdExecutor([BIN_CP, "-a", "/etc.defaults/VERSION", SZD_OLDPATCH_DIR]):
        sdk.log("fail to copy VERSION")
        return False
    if not sdk.CmdExecutor([BIN_CP, "-a", "/.system_info", SZD_OLDPATCH_DIR]):
        sdk.log("fail to copy /.system_info")
        return False

    # archive old patch info
    if not sdk.CmdExecutor([BIN_TAR, "-cf", SZF_OLDPATCH_PATH, "-C", SZF_ARCHIVE_DIR, ".old_patch_info"]):
        sdk.log("fail to copy /.system_info")
        return False

    return True


def CheckSystemPartitionSize():
    # check if system partition size is less than 50MB
    disk = os.statvfs('/')
    free_mb = (disk.f_bfree * disk.f_bsize) / (1024*1024)
    return free_mb > 50


def PgsqlDump():
    '''
    /bin/su postgres -c "/usr/bin/pg_dumpall --disable-triggers -f /var/services/pgsql/pgsql-32bit.sql"
    mv /var/services/pgsql/pgsql-32bit.sql /syno_migration/pgsql-32bit.sql
    '''
    dump_path = "/var/services/pgsql/pgsql-32bit.sql"
    dump_cmd = "/usr/bin/pg_dumpall --disable-triggers -f {}".format(dump_path)
    if not sdk.CmdExecutor(["/bin/su", "postgres", "-c", dump_cmd]):
        sdk.log("fail to invoke pg_dumpall")
        return False

    if not sdk.CmdExecutor([BIN_MV, dump_path, SZF_ARCHIVE_DIR+'pgsql-32bit.sql']):
        sdk.log("fail to mv pgsql-32bit.sql")
        return False

    return True


def NeedPgsqlMigrate():
    # platform is 32bits (only alpine now) and volumes exist
    if sdk.isAlpinePlatform() and 0 != len(sdk.VolumeList()):
        return True


if __name__ == '__main__':

    if not SystemConfigArchive():
        sdk.log("fail to archive system config")
        if not CheckSystemPartitionSize():
            exit(2)
        exit(1)

    if not SystemInfoArchive():
        sdk.log("fail to archive system info")
        if not CheckSystemPartitionSize():
            exit(2)
        exit(1)

    # handle pgsql 32/64 bits migration issue
    if NeedPgsqlMigrate() and not PgsqlDump():
        sdk.log("fail to dump pgsql")
        if not CheckSystemPartitionSize():
            exit(2)
        exit(1)

    exit(0)

