#!/bin/python

import os
import stat
import json
import subprocess

import sdk
import drbdutil

# binary path definition
BIN_LOSETUP="/sbin/losetup"
BIN_DD="/bin/dd"
BIN_MKNOD="/bin/mknod"

# kernal module definition
KO_DRBD="/lib/modules/drbd.ko"
KO_LOOP="/lib/modules/loop.ko"

# drbd config relative definition
DIR_SYNO_DRBD="/usr/syno/etc/synomigration/drbd.d/"
CFG_SYNO_DRBD=DIR_SYNO_DRBD+"{}.res"

DEV_DRBD_PREFIX="/dev/drbd-"
DEV_LOOP_PREFIX="/dev/loop."
DEV_BACK_DIR="/var/lib/drbd_md/"
DEV_BACK_PREFIX="/var/lib/drbd_md/drbd_md."

# migration config relative definition
CFG_SYNO_MIGRATION="/usr/syno/etc/synomigration/info.cfg"

SZK_SYNO_MIGRATION_MIGRATION_ROLE="migration_role"
SZK_SYNO_MIGRATION_LOCAL_IP="local_ip"
SZK_SYNO_MIGRATION_REMOTE_IP="remote_ip"
SZK_SYNO_MIGRATION_MIGRATION_SETTING="migration_setting"
SZK_SYNO_MIGRATION_VOLUME_MAPPING="volume_mapping"
SZK_SYNO_MIGRATION_SYNC_ID="id"
SZK_SYNO_MIGRATION_LOCAL_DISK="local_disk"
SZK_SYNO_MIGRATION_REMOTE_DISK="remote_disk"
SZK_SYNO_MIGRATION_DEVICE_TYPE="type"
SZK_SYNO_MIGRATION_FS_TYPE="fs_type"
SZK_SYNO_MIGRATION_SIZE="size"

SZK_SYNO_MIGRATION_MNT_POINT="mnt_point"
SZK_SYNO_MIGRATION_MNT_OPT="mnt_opt"

SYNO_MIGRATION_ROLE_SRC="source"
SYNO_MIGRATION_ROLE_DST="destination"

def CreateFolderIfNotExisted(folder_path):
    if os.path.isdir(folder_path):
        return True

    if os.path.exists(folder_path):
        # existed but not a folder
        return False

    os.mkdir(folder_path)
    return True


def ReferNameGetBySpace(space_path):
    '''
    Get reference name for drbd from space path, internal use
        ex. /dev/md2 => "md2"
        ex. /dev/vg1/volume_1 => "vg1-volume_1"
        ex. /dev/vg2/iscsi_0 => "vg2-iscsi_0"
    '''

    if len(space_path) < 5 or space_path[:5] != "/dev/":
        return ''

    refer_name = space_path[5:].replace("/", "-");

    return refer_name


def MigrationInfoGetByDiskName(local_disk_name):

    # read config
    mgr_cfg = open(CFG_SYNO_MIGRATION, mode='r')
    if not mgr_cfg:
        sdk.log("fail to open migration config file: {}".format(CFG_SYNO_MIGRATION))
        return {}

    # load into json
    mgr_json = json.load(mgr_cfg)
    if not mgr_cfg:
        sdk.log("fail to load migration json: {}".format(CFG_SYNO_MIGRATION))
        return {}

    # find disk mapping by local disk name
    drbd_info = {}
    disk_mapping = mgr_json[SZK_SYNO_MIGRATION_MIGRATION_SETTING][SZK_SYNO_MIGRATION_VOLUME_MAPPING]
    for entry in disk_mapping:
        if entry['local_disk'] == local_disk_name:
            drbd_info['local_ip'] = mgr_json[SZK_SYNO_MIGRATION_LOCAL_IP]
            drbd_info['remote_ip'] = mgr_json[SZK_SYNO_MIGRATION_REMOTE_IP]

            drbd_info['migration_id'] = entry[SZK_SYNO_MIGRATION_SYNC_ID]
            drbd_info['local_disk'] = entry[SZK_SYNO_MIGRATION_LOCAL_DISK]
            drbd_info['remote_disk'] = entry[SZK_SYNO_MIGRATION_REMOTE_DISK]
            drbd_info['size'] = entry[SZK_SYNO_MIGRATION_SIZE]
            break

    if not drbd_info:
        sdk.log("fail to find config by disk name: {}".format(local_disk_name))

    mgr_cfg.close()

    return drbd_info


def VolumeInfoGetBySpaceRefer(space_refer):

    # read config
    mgr_cfg = open(CFG_SYNO_MIGRATION, mode='r')
    if not mgr_cfg:
        sdk.log("fail to open migration config file: {}".format(CFG_SYNO_MIGRATION))
        return {}

    # load into json
    mgr_json = json.load(mgr_cfg)
    if not mgr_cfg:
        sdk.log("fail to load migration json: {}".format(CFG_SYNO_MIGRATION))
        return {}

    # find disk mapping by local disk name
    vol_info = {}
    disk_mapping = mgr_json[SZK_SYNO_MIGRATION_MIGRATION_SETTING][SZK_SYNO_MIGRATION_VOLUME_MAPPING]
    for entry in disk_mapping:
        if entry[SZK_SYNO_MIGRATION_MNT_POINT] == space_refer:

            # FIXME: handle invalid configure
            vol_info['type'] = entry[SZK_SYNO_MIGRATION_DEVICE_TYPE]
            vol_info['dev_path'] = entry[SZK_SYNO_MIGRATION_LOCAL_DISK]
            vol_info['mnt_opt'] = entry[SZK_SYNO_MIGRATION_MNT_OPT]

            # this field is used by quota, which block lun doesn't have it
            if SZK_SYNO_MIGRATION_FS_TYPE in entry:
                vol_info['fs_type'] = entry[SZK_SYNO_MIGRATION_FS_TYPE]

            break

    if not vol_info:
        sdk.log("fail to find config by mount point: {}".format(space_refer))

    mgr_cfg.close()

    return vol_info


def MigrationRoleGet():

    # read config
    mgr_cfg = open(CFG_SYNO_MIGRATION, mode='r')
    if not mgr_cfg:
        sdk.log("fail to open migration config file: {}".format(CFG_SYNO_MIGRATION))
        return ""

    # load into json
    mgr_json = json.load(mgr_cfg)
    if not mgr_cfg:
        sdk.log("fail to load migration json {}".format(CFG_SYNO_MIGRATION))
        return ""

    # get migration role
    role = mgr_json[SZK_SYNO_MIGRATION_MIGRATION_ROLE]
    if not role:
        sdk.log("fail to get migration role from {}".format(CFG_SYNO_MIGRATION))
        return ""

    return role


def DrbdConfigCreate(disk_name, port):
    '''
    Create drbd config (/usr/syno/etc/synomigration/drbd.d/$MIGRATION_ID.res)
    by migration config (/usr/syno/etc/synomigration/info.cfg)
    '''

    # read migration config
    drbd_info = MigrationInfoGetByDiskName(disk_name)
    if not drbd_info:
        sdk.log("fail to read migration config"+ disk_name)
        return False

    print "Migration ID: {}".format(drbd_info['migration_id'])
    print "Local IP: {}".format(drbd_info['local_ip'])
    print "Local disk: {}".format(drbd_info['local_disk'])
    print "Remote IP: {}".format(drbd_info['remote_ip'])
    print "Remote disk: {}".format(drbd_info['remote_disk'])

    # write drbd config
    local_refer = ReferNameGetBySpace(drbd_info['local_disk'])
    if not local_refer:
        sdk.log("fail to get refer name from disk"+ drbd_info['local_disk'])
        return False
    remote_refer = ReferNameGetBySpace(drbd_info['remote_disk'])
    if not remote_refer:
        sdk.log("fail to get refer name from disk"+ drbd_info['remote_disk'])
        return False

    '''
    resource $MIGRATION_ID {
        floating $LOCAL_IP : 7400 {
            device /dev/drbd-$LOCAL_REFER ;
            disk $LOCAL_DISK ;
            meta-disk /dev/loop.$LOCAL_REFER ;
        }
        floating $REMOTE_IP : 7400 {
            device /dev/drbd-$REMOTE_REFER ;
            disk $REMOTE_DISK ;
            meta-disk /dev/loop.$REMOTE_REFER ;
        }
    }
    '''

    # remove if existed
    drbd_file_path = CFG_SYNO_DRBD.format(drbd_info['migration_id'])
    if os.path.isfile(drbd_file_path):
        os.unlink(drbd_file_path)

    # create new config
    CreateFolderIfNotExisted(DIR_SYNO_DRBD)
    sdk.log("[INFO] create drbd config: {}".format(drbd_file_path))
    with open(drbd_file_path, 'w+') as file:
        file.write('resource {} {{\n'.format(drbd_info['migration_id']))

        file.write('    floating {} : {} {{\n'.format(drbd_info['local_ip'], port))
        file.write('        device {0}{1} ;\n'.format(DEV_DRBD_PREFIX, local_refer))
        file.write('        disk {} ;\n'.format(drbd_info['local_disk']))
        file.write('        meta-disk {0}{1} ;\n'.format(DEV_LOOP_PREFIX, local_refer))
        file.write('    }\n')

        file.write('    floating {} : {} {{\n'.format(drbd_info['remote_ip'], port))
        file.write('        device {0}{1} ;\n'.format(DEV_DRBD_PREFIX, local_refer))
        file.write('        disk {} ;\n'.format(drbd_info['local_disk']))
        file.write('        meta-disk {0}{1} ;\n'.format(DEV_LOOP_PREFIX, local_refer))

        file.write('    }\n')

        file.write('}\n')

    if not os.path.isfile(drbd_file_path):
        sdk.log("fail to create drbd config")
        return False

    return True


def DrbdDeviceCreate(refer_name, node):
    '''
    Create drbd device and a link to refer it (/dev/drbd-$REFER)

        drbd device:    mknod /dev/drbd0 b 147 0
        drbd alias:     ln -s /dev/drbd0 /dev/drbd-$REFER
    '''

    # reuse existed
    drbd_device = "{0}{1}".format(DEV_DRBD_PREFIX, refer_name)

    if os.path.isfile(drbd_device) and stat.S_ISBLK(os.stat(drbd_device).st_mode):
        return True

    # create device
    cmd_arg = [BIN_MKNOD, drbd_device, "b", "147", node]
    output, err = subprocess.Popen(cmd_arg, stdout=subprocess.PIPE).communicate()

    if not stat.S_ISBLK(os.stat(drbd_device).st_mode):
        sdk.log("fail to create drbd device: {}".format(drbd_device))
        return False

    return True


def LoopDeviceFreeGet():
    cmd_arg = [BIN_LOSETUP, "-f"]
    output, err = subprocess.Popen(cmd_arg, stdout=subprocess.PIPE).communicate()
    return output.rstrip()


def LoopDeviceCreate(refer_name, device_size):
    '''
    Create a loop device and a link to refer it (/dev/loop.$REFER)

        loop device:    dd if=/dev/zero of=/var/lib/drbd_md/drbd_md.$REFER bs=1M count=10
                        losetup -d /dev/loop0
                        losetup -d /var/lib/drbd_md/drbd_md.$REFER
                        losetup /dev/loop0 /var/lib/drbd_md/drbd_md.$REFER
        loop alias:     ln -s /dev/loop0 /dev/loop.$REFER
    '''

    if not refer_name:
        sdk.log("invalid refer name")
        return False

    # compose device name
    refer_link = "{0}{1}".format(DEV_LOOP_PREFIX, refer_name)
    backing_file = "{0}{1}".format(DEV_BACK_PREFIX, refer_name)
    loop_device = LoopDeviceFreeGet()
    sdk.log("[INFO] LoopDeviceFreeGet: {}".format(loop_device))
    if not loop_device:
        sdk.log("fail to get free loop device: {}".format(loop_device))
        return False

    # prepare backing file folder
    if not CreateFolderIfNotExisted(DEV_BACK_DIR):
        sdk.log("fail to create backing file folder: {}".format(DEV_BACK_DIR))
        return False

    # create backing file
    sdk.log("[INFO] create backing file: {}".format(backing_file))

    backing_file_size_mb = (device_size / (1024 * 1024 * 2097152)) + 2
    cmd_arg = [BIN_DD, "if=/dev/zero", "of={0}".format(backing_file), "bs=1M", "count={0}".format(backing_file_size_mb)]
    output, err = subprocess.Popen(cmd_arg, stdout=subprocess.PIPE).communicate()

    mode = os.stat(backing_file).st_mode
    if not stat.S_ISREG(mode):
        sdk.log("fail to create backing device: {}".format(backing_device))
        return False

    # loop device setup
    #cmd_arg = [BIN_LOSETUP, "-d", loop_device]
    #output, err = subprocess.Popen(cmd_arg, stdout=subprocess.PIPE).communicate()

    #cmd_arg = [BIN_LOSETUP, "-d", backing_file]
    #output, err = subprocess.Popen(cmd_arg, stdout=subprocess.PIPE).communicate()

    cmd_arg = [BIN_LOSETUP, loop_device, backing_file]
    output, err = subprocess.Popen(cmd_arg, stdout=subprocess.PIPE).communicate()

    # make alias
    sdk.log("[INFO] create loop alias: {0} -> {1}".format(refer_link, loop_device))
    if os.path.islink(refer_link):
        os.unlink(refer_link)
    cmd_arg = ["/bin/ln", "-s", loop_device, refer_link]
    output, err = subprocess.Popen(cmd_arg, stdout=subprocess.PIPE).communicate()

    return True


def DrbdDeviceSetup(resource_name):

    drbdutil.DrbdDown(resource_name)
    drbdutil.DrbdDownExtra(resource_name)

    # drbdadm create-md $resource
    if not drbdutil.DrbdCreateMD(resource_name):
        sdk.log("fail to DrbdCreateMD: {}".format(resource_name))
        return False

    # drbdadm new-resource $resource
    if not drbdutil.DrbdNewResource(resource_name):
        sdk.log("fail to DrbdNewResource: {}".format(resource_name))
        return False

    # drbdadm sh-new-minor $resource
    if not drbdutil.DrbdShNewMinor(resource_name):
        sdk.log("fail to DrbdShNewMinor: {}".format(resource_name))
        return False

    # drbdadm attach $resource
    if not drbdutil.DrbdAttach(resource_name):
        sdk.log("fail to DrbdAttach: {}".format(resource_name))
        return False

    # drbdadm primary $resource
    migration_role = MigrationRoleGet()

    if SYNO_MIGRATION_ROLE_SRC == migration_role:
        if not drbdutil.DrbdPrimary(resource_name):
            sdk.log("fail to DrbdPrimary: {}".format(resource_name))
            return False

        if not drbdutil.DrbdConnect(resource_name):
            sdk.log("fail to DrbdConnect: {}".format(resource_name))
            return False

    return True


def ResourceCreate(volume_path, space_path, port):
    '''
    Prepare following drbd resources
         - resource:               /etc/drbd.d/vg1-volume_1.res

         - drbd device:            mknod /dev/drbd-vg1-volume_1 b 147 0
         - drbd alias:             ln -s /dev/drbd-vg1-volume_1 /dev/drbd0

         - loop device:            losetup /dev/loop0 /var/lib/drbd_md/drbd_md.vg1-volume_1
         - loop alias:             ln -s /dev/loop.vg1-volume_1 /dev/loop0

    And setup drbd device afterward
         - drbdadm create-md $resource
         - drbdadm new-resource $resource
         - drbdadm sh-new-minor $resource
         - drbdadm attach $resource
         - drbdadm primary $resource
    '''

    # initialization
    sdk.KernelModuleInsert(KO_DRBD)
    sdk.KernelModuleInsert(KO_LOOP)

    # get refer name
    if not space_path:
        sdk.log("invalid space path: {}".format(space_path))
        return False

    # read refer name in config
    refer_name = ReferNameGetBySpace(space_path)
    if not refer_name:
        sdk.log("fail to get refer name: {}".format(space_path))
        return False

    # get resource name
    drbd_info = MigrationInfoGetByDiskName(space_path)
    resource_name = drbd_info['migration_id']
    device_size = drbd_info['size']

    # 1. create drbd config
    if not DrbdConfigCreate(space_path, port):
        sdk.log("fail to create drbd config: {}".format(space_path))
        return False

    # 2. create drbd device
    if not DrbdDeviceCreate(refer_name, str(int(port)-7400)):
        sdk.log("fail to create drbd device: {}".format(refer_name))
        return False

    # 3. create loop device
    if not LoopDeviceCreate(refer_name, device_size):
        sdk.log("fail to create drbd loop: {}".format(refer_name))
        return False

    # 4. setup drbd device
    if not DrbdDeviceSetup(resource_name):
        sdk.log("fail to setup drbd device: {}".format(resource_name))
        return False


    drbd_device_name = "{0}{1}".format(DEV_DRBD_PREFIX, refer_name)

    return drbd_device_name


############################################################################
############################################################################
############################################################################
############################################################################
############################################################################

def DrbdDeviceTeardown(resource_name):

    if not drbdutil.DrbdPauseSync(resource_name):
        sdk.log("fail to DrbdPauseSync: {}".format(resource_name))
        return False

    if not drbdutil.DrbdSecondary(resource_name):
        sdk.log("fail to DrbdSecondary: {}".format(resource_name))
        return False

    if not drbdutil.DrbdDisconnect(resource_name):
        sdk.log("fail to DrbdDisconnect: {}".format(resource_name))
        return False

    if not drbdutil.DrbdDown(resource_name):
        sdk.log("fail to DrbdDown: {}".format(resource_name))
        return False

    if not drbdutil.DrbdDownExtra(resource_name):
        sdk.log("fail to DrbdDownExtra: {}".format(resource_name))
        return False

    return True


def LoopDeviceFree(refer_name):
    '''
     - refer name:         /dev/loop.vg1-volume_1
     - backing file:       /var/lib/drbd_md/drbd_md.vg1-volume_1
     - loop device:        /dev/loop0
    '''

    # compose device names
    refer_link = "{0}{1}".format(DEV_LOOP_PREFIX, refer_name)
    backing_file = "{0}{1}".format(DEV_BACK_PREFIX, refer_name)
    loop_device = os.path.realpath(refer_link)

    # free loop device (?)
    sdk.log("[INFO] free loop device: {}".format(loop_device))
    cmd_arg = [BIN_LOSETUP, "-d", loop_device]
    output, err = subprocess.Popen(cmd_arg, stdout=subprocess.PIPE).communicate()

    # delete backing file
    sdk.log("[INFO] delete backing file: {}".format(backing_file))
    if os.path.isfile(backing_file):
        os.unlink(backing_file)

    # delete alias
    sdk.log("[INFO] delete device alias: {}".format(refer_link))
    if os.path.islink(refer_link):
        os.unlink(refer_link)

    return True


def DrbdDeviceDelete(refer_name):
    '''
     - drbd device:            /dev/drbd-vg1-volume_1
    '''

    # delete drbd device
    drbd_device = "{0}{1}".format(DEV_DRBD_PREFIX, refer_name)

    sdk.log("[INFO] delete drbd device: {}".format(drbd_device))
    if not os.path.exists(drbd_device):
        return True

    #if not os.unlink(drbd_device):
    #    sdk.log("fail to delete drbd drvice {}".format(drbd_device))
    #    return False

    os.unlink(drbd_device)

    return True


def DrbdConfigDelete(disk_name):

    # read migration config
    drbd_info = MigrationInfoGetByDiskName(disk_name)
    if not drbd_info:
        sdk.log("fail to read migration config {}".format(disk_name))
        return False

    # remove config file
    drbd_file_path = CFG_SYNO_DRBD.format(drbd_info['migration_id'])

    sdk.log("[INFO] delete drbd config: {}".format(drbd_file_path))
    if not os.path.isfile(drbd_file_path):
        return True

    #if not os.unlink(drbd_file_path):
    #    sdk.log("fail to delete drbd config {}".format(drbd_file_path))
    #    return False

    os.unlink(drbd_file_path)

    return True


def ResourceRemove(volume_path, space_path):
    '''
    Teardown drbd device afterward
         - drbdadm pause-sync $resource
         - drbdadm secondary $resource
         - drbdadm disconnect $resource
         - drbdadm down $resource
         - drbdadm down-extra $resource

    And remove following drbd resources
         - free loop device:     losetup -d /dev/loop0 (?)
         - delete drbd md:         rm /var/lib/drbd_md/drbd_md.vg1-volume_1
         - delete loop alias:      rm /dev/loop.vg1-volume_1

         - delete drbd device:     rm /dev/drbd-vg1-volume_1
         - delete drbd alias:      rm /dev/drbd0

         - delete resource:        rm /etc/drbd.d/vg1-volume_1.res
    '''

    # get refer name
    if not space_path:
        sdk.log("invalid space path: {}".format(space_path))
        return False

    refer_name = ReferNameGetBySpace(space_path)
    if not refer_name:
        sdk.log("fail to get refer name: {}".format(space_path))
        return False

    # get resource name
    drbd_info = MigrationInfoGetByDiskName(space_path)
    resource_name = drbd_info['migration_id']

    # 1. teardown drbd device
    if not DrbdDeviceTeardown(resource_name):
        sdk.log("fail to teardown drbd device: {}".format(resource_name))
        return False

    # 2. free loop device
    if not LoopDeviceFree(refer_name):
        sdk.log("fail to free loop device: {}".format(refer_name))
        return False

    # 3. delete drbd device
    if not DrbdDeviceDelete(refer_name):
        sdk.log("fail to delete drbd device: {}".format(refer_name))
        return False

    # 4. delete drbd config
    if not DrbdConfigDelete(space_path):
        sdk.log("fail to delete drbd config: {}".format(space_path))
        return False

    return True


# vim: set ts=4 sts=4 sw=4 noet:
