#!/bin/python

import ctypes
import subprocess
import json
import os
import time

BIN_SYNO_GETKEYVALUE = "/bin/get_key_value"
BIN_CAT = "/bin/cat"
BIN_SED = "/bin/sed"
BIN_CP = "/bin/cp"
BIN_MV = "/bin/mv"
BIN_JQ = "/bin/jq"
BIN_MOUNT = "/bin/mount"
BIN_UMOUNT = "/bin/umount"
BIN_SYNO_USER = "/usr/syno/sbin/synouser"
BIN_SYNO_GROUP = "/usr/syno/sbin/synogroup"
BIN_SYNO_ISCSI = "/usr/syno/bin/synoiscsiep"
BIN_SYNO_SERVICE = "/usr/syno/sbin/synoservice"
BIN_SYNO_PKG = "/usr/syno/bin/synopkg"
BIN_SYNO_LOG = "/usr/syno/bin/synologset1"

BIN_INSMOD="/sbin/insmod"
BIN_RMMOD="/sbin/rmmod"

# TODO: compatibility test for libsynosdk
LIB_SYNO_SDK = "/lib/libsynosdk.so.6"

LIB_SYNO_SERVICE = "/lib/libsynoservice.so"
LIB_SYNO_STORAGECORE = "/lib/libsynostoragecore.so"
LIB_SYNO_LOCALACCOUNT = "/lib/libsynolocalaccount.so"

SZF_DEV_ARG = "/usr/syno/etc/synomigration/dev_arg"

SYNOServiceStart = 0
SYNOServiceStop = 1

DSM_52_START_VERSION =  5100
DSM_60_START_VERSION =  7100
DSM_61_START_VERSION = 14600
DSM_62_START_VERSION = 22100
DSM_70_START_VERSION = 40000

SZ_SYNO_PAUSE_REASON = "Synology Migration Assistant"

def log(msg):
    # print msg
    import syslog
    syslog.syslog(syslog.LOG_ERR, msg)


def CmdExecutor(cmd_arg):
    try:
        retcode = subprocess.call(cmd_arg, stdout=subprocess.PIPE)

        if retcode < 0:
            log("{0} was terminated by signal: {1}".format(cmd_arg[0], -retcode))
            return False
        elif retcode != 0 and retcode != 10:
            log("{0} was returned: {1}".format(cmd_arg[0], retcode))
            return False
        else:
            return True
    except OSError as e:
        log("{0} execution failed: {1}".format(cmd_arg[0], e))
        return False


def SynoLog(msg):
    cmd_arg=[BIN_SYNO_LOG, "sys", "info", "0x11100000", msg]
    CmdExecutor(cmd_arg)


def KernelModuleIsLoaded(module):
    return os.path.exists("/sys/module/" + module)


def KernelModuleInsert(module):
    cmd_arg=[BIN_INSMOD, module]
    return CmdExecutor(cmd_arg)


def KernelModuleRemove(module):
    cmd_arg=[BIN_RMMOD, module]
    return CmdExecutor(cmd_arg)


def SynoModelGet():
    cmd_arg = [BIN_SYNO_GETKEYVALUE, '/etc.defaults/synoinfo.conf', 'unique']
    output, err = subprocess.Popen(cmd_arg, stdout=subprocess.PIPE).communicate()
    return output.rstrip()


def SynoVersionGet():
    cmd_arg = [BIN_SYNO_GETKEYVALUE, '/etc.defaults/VERSION', 'buildnumber']
    output, err = subprocess.Popen(cmd_arg, stdout=subprocess.PIPE).communicate()
    return output.rstrip()


def IsHomeEnable():
    cmd_arg = [BIN_SYNO_GETKEYVALUE, '/etc/synoinfo.conf', 'userHomeEnable']
    output, err = subprocess.Popen(cmd_arg, stdout=subprocess.PIPE).communicate()
    return ("yes" == output.rstrip())


def SynoFeasibilityCheck():
    '''
    SYNOFeasibilityCheck(eFType, CTYPE_HARD, &pslTaskMsg, args..., NULL)
    '''

    # TODO: implementation

    return True


def SLIBServicePauseByReason(service, reason):
    '''
    int SLIBServicePauseByReason(const char *szServ, const char *szReason)
    '''

    if not service or not reason:
        log("invalid SLIBServicePauseByReason argument: {0} {1}".format(service, reason))
        return False

    # COMPATIBILITY: service might not existed in previous DSM
    service_cfg = "/usr/syno/etc/synoservice.d/{}.cfg".format(service)
    if not os.path.isfile(service_cfg):
        return True

    libsynosdk = ctypes.cdll.LoadLibrary(LIB_SYNO_SDK)
    if not libsynosdk:
        log("fail to load library: {}".format(LIB_SYNO_SDK))
        return False

    libsynosdk.SLIBServicePauseByReason.restype = ctypes.c_int

    try:
        ret = libsynosdk.SLIBServicePauseByReason(service, reason)
        if 0 > ret:
            log("result of SLIBServicePauseByReason: {0}".format(ret))
            return False
    except Exception:
        log("fail to invoke SLIBServicePauseByReason")
        return False

    return True


def SLIBServiceResumeByReason(service, reason):
    '''
    int SLIBServiceResumeByReason(const char *szServ, const char *szReason)
    '''

    if not service or not reason:
        log("invalid SLIBServiceResumeByReason argument: {0} {1}".format(service, reason))
        return False

    # COMPATIBILITY: service might not existed in previous DSM
    service_cfg = "/usr/syno/etc/synoservice.d/{}.cfg".format(service)
    if not os.path.isfile(service_cfg):
        return True

    libsynosdk = ctypes.cdll.LoadLibrary(LIB_SYNO_SDK)
    if not libsynosdk:
        log("fail to load library: {}".format(LIB_SYNO_SDK))
        return False

    libsynosdk.SLIBServiceResumeByReason.restype = ctypes.c_int

    try:
        ret = libsynosdk.SLIBServiceResumeByReason(service, reason)
        if 0 > ret:
            log("result of SLIBServiceResumeByReason: {0}".format(ret))
            return False
    except Exception:
        log("fail to invoke SLIBServiceResumeByReason")
        return False

    return True


def SynoServiceControl(action):
    '''
    SYNOServiceOnOffALL(SLIB_ACTION_START/SLIB_ACTION_STOP)
    '''

    libsynoservice = ctypes.cdll.LoadLibrary(LIB_SYNO_SERVICE)
    if not libsynoservice:
        log("fail to load library: {}".format(LIB_SYNO_SERVICE))
        return False

    libsynoservice.SYNOServiceOnOffALL.restype = ctypes.c_int

    try:
        ret = libsynoservice.SYNOServiceOnOffALL(action)
        if 0 > ret:
            log("result of SYNOServiceOnOffALL: {0}".format(ret))
            return False
    except:
        log("fail to invoke SYNOServiceOnOffALL")
        return False

    return True


def SynoServiceExecutor(service, operation):
    cmd = [BIN_SYNO_SERVICE, operation, service]
    try:
        retcode = subprocess.call(cmd, stdout=subprocess.PIPE)
        return 0 == retcode
    except:
        return False


def SynoServiceResumeByReason(reason):
    cmd = [BIN_SYNO_SERVICE, "--resume-all", reason]
    try:
        retcode = subprocess.call(cmd, stdout=subprocess.PIPE)
        return 0 == retcode
    except:
        return False


def FixInvalidJSON(cfg_file):
    if not os.path.exists(cfg_file):
        return True

    BIN_JSON_CLEANER="/usr/syno/etc/synomigration/scripts/python/space/json_cleaner.py"
    if not CmdExecutor([BIN_JSON_CLEANER, cfg_file]):
        log("fail to clean invalid json config: {}".format(cfg_file))
        return False

    return True

def loadSynoServiceConfig(service, override=False):
    '''
    FIXME: some json file format is incorrect (e.g. /usr/syno/etc/synoservice.d/synotunnel.cfg)
    '''

    srv_dir = "synoservice.override" if override else "synoservice.d"
    srv_path = "/usr/syno/etc/{0}/{1}.cfg".format(srv_dir, service)

    # check file existency
    if not os.path.isfile(srv_path):
        return {}

    # open file
    srv_cfg = open(srv_path, mode='r')
    if not srv_cfg:
        log("fail to open migration config file: {}".format(srv_path))
        return {}

    # load into json
    try:
        srv_json = json.load(srv_cfg)
    except ValueError:
        pass
        try:
            FixInvalidJSON(srv_path)
            srv_json = json.load(srv_cfg)
        except ValueError:
            return {}

    if not srv_json:
        log("fail to load migration json: {}".format(srv_path))
        return {}

    return srv_json


def IsServiceNeedToStop(service):

    control_list = [
        "samba",
        "atalk",
        "tftp",
        "sftp",
        "ftpd",
        "ftpd-ssl",
        "rsyncd",
        "nfsd",
        "iscsitrg"
    ]

    if not service in control_list:
        return False

    svr_json = loadSynoServiceConfig(service, False)
    svr_or_json = loadSynoServiceConfig(service, True)

    '''
    if not 'user_controllable' in svr_json or 'yes' != svr_json['user_controllable']:
        return False

    if 'is_package' in svr_json and 'yes' == svr_json['is_package']:
        return False
    '''

    # override is on
    if 'auto_start' in svr_or_json and 'yes' == svr_or_json['auto_start']:
        return True

    # default is on
    if 'auto_start' in svr_json and 'yes' == svr_json['auto_start']:
        return True

    return False


def DisableAlliSCSITarget():
    '''
    /usr/syno/bin/synowebapi --exec api=SYNO.Storage.CGI.Storage method=load_info (data.iscsiTargets[].tid)
    /usr/syno/bin/synowebapi --exec api=SYNO.Core.Storage.iSCSITargets method=disable tid=$TID
    '''

    j_data = execWebAPI('SYNO.Storage.CGI.Storage', 1, 'load_info')
    if not j_data or not j_data['success']:
        return False

    if not j_data['data'] or not j_data['data']['iscsiTargets']:
        return False

    for target in j_data['data']['iscsiTargets']:
        j_data = execWebAPI('SYNO.Core.Storage.iSCSITargets', 1, 'disable', tid=target['tid'])
        if not j_data or not j_data['success']:
            log("fail to disable iSCSI targt {}".format(target['tid']))

    return True


def SynoDisableInternalServices():
    for cfg_file in os.listdir("/usr/syno/etc/synoservice.d/"):
        if not cfg_file.endswith(".cfg"):
            continue

        service_name = cfg_file[:-4]
        if not IsServiceNeedToStop(service_name):
            continue

        # stop service
        log("stop service: {0}".format(service_name))
        if "iscsitrg" == service_name:
            if not DisableAlliSCSITarget():
                log("fail to disable all iSCSI target")
        else:
            if not SynoServiceExecutor(service_name, "--disable"):
                log("fail to disable service: {0}".format(service_name))

    return True


def SynoDisablePackages():
    '''
    for each package, execute /usr/syno/bin/synopkg stop <package>
    '''

    skip_list = ["SynoFinder"]

    for pkg in os.listdir("/var/packages/"):
        if pkg in skip_list:
            continue

        if not CmdExecutor([BIN_SYNO_PKG, "stop", pkg]):
            log("fail to stop package: {0}".format(pkg))

    return True


def SynoServiceDisableAll():
    '''
    FIXME:
        We want to disable service from paused status.
        The efficient solution is to clear paused reason directly without resume service.
        (include: Synology Migration Assistant / ErrNetService)

        But we don't have such API. So I resume the service and disable it for now.
    '''

    # disable package
    if not SynoDisablePackages():
        log("fail to disable packages")

    # disable internal services
    if not SynoDisableInternalServices():
        log("fail to disable internal service")

    return True


def isAlpinePlatform():
    model = SynoModelGet()
    return "alpine" in model


def SynoIscsiControl(start):
    '''
    We cannot ensure there is no access by stopping iSCSI by --stop/--start.
    We need to stop service entirely by --oil/--oilio.
    '''

    # it might fail to stop iSCSI if vhost_net is loaded
    if KernelModuleIsLoaded("vhost_net"):
        if not KernelModuleRemove("vhost_net"):
            log("fail to remove kernel module: vhost_net")

    subcmd = ""
    if start:
        subcmd = "--oilio"
    else:
        subcmd = "--oil"

    ret = CmdExecutor([BIN_SYNO_ISCSI, subcmd])

    # WORKAROUND: on apline 6.0/6.1, it might fail to start iSCSI due to continuous memory allocation failure
    if start and int(SynoVersionGet()) < DSM_62_START_VERSION and isAlpinePlatform():
        log("fail to start iSCSI service due to continuous memory allocation")
        return True
    else:
        return ret


def SynoServiceStartAll():
    if not SynoIscsiControl(True):
        return False

    if not SynoServiceControl(SYNOServiceStart):
        return False

    # WORKAROUND: s2s_daemon will not be stopped, but it will access volume
    if not SLIBServiceResumeByReason("s2s_daemon", SZ_SYNO_PAUSE_REASON):
        return False

    # WORKAROUND: sshd won't be stopped, but it will access volume if home is enabled
    if not SLIBServiceResumeByReason("ssh-shell", SZ_SYNO_PAUSE_REASON):
        return False

    # WORKAROUND: synologanalyzer might not be stopped sometimes, but it will access volume
    if not SLIBServiceResumeByReason("synologanalyzer", SZ_SYNO_PAUSE_REASON):
        return False

    # WORKAROUND: fixed a known issue that winbind will access volume (DSM#115926)
    if not SLIBServiceResumeByReason("winbindd", SZ_SYNO_PAUSE_REASON):
        return False

    return True


def SynoServiceStopAll():

    # WORKAROUND: package user home might be enabled somehow
    if UserHomeCheck("MigrationAssistant") and not sdk.UserHomeDisable("MigrationAssistant"):
        log("fail to disable package user home")
        return False

    # WORKAROUND: s2s_daemon will not be stopped, but it will access volume
    if not SLIBServicePauseByReason("s2s_daemon", SZ_SYNO_PAUSE_REASON):
        return False

    # WORKAROUND: sshd won't be stopped, but it will access volume if home is enabled
    if not SLIBServicePauseByReason("ssh-shell", SZ_SYNO_PAUSE_REASON):
        return False

    # WORKAROUND: synologanalyzer might not be stopped sometimes, but it will access volume
    if not SLIBServicePauseByReason("synologanalyzer", SZ_SYNO_PAUSE_REASON):
        return False

    # WORKAROUND: fixed a known issue that winbind will access volume (DSM#115926)
    if not SLIBServicePauseByReason("winbindd", SZ_SYNO_PAUSE_REASON):
        return False

    if not SynoServiceControl(SYNOServiceStop):
        return False

    if not SynoIscsiControl(False):
        return False

    return True


def SynoMountInfoGetByVolume(volume_path):
    '''
    Get disk name by volume path by parsing fstab
        - cat /etc/fstab | grep /volume1
        - /dev/md2 /volume1 ext4 usrjquota=aquota.user,grpjquota=aquota.group,jqfmt=vfsv0,nomblk_io_submit,synoacl,relatime 0 0
    '''

    cmd_arg = [BIN_CAT, "/etc/fstab"]
    fstab = subprocess.Popen(cmd_arg, stdout=subprocess.PIPE)

    mount_line = ''
    while True:
        line = fstab.stdout.readline()

        if volume_path in line:
            mount_line = line
            break
        elif '' == line:
            break

    if not mount_line:
        log("fail to find device path of {}".format(volume_path))
        return {}

    mount_info = {}
    mount_array = mount_line.split(" ")
    if not (6 == len(mount_array)):
        return {}

    mount_info['device_path'] = mount_array[0]
    mount_info['mount_path'] = mount_array[1]
    mount_info['file_system'] = mount_array[2]
    mount_info['mount_option'] = mount_array[3]

    return mount_info


def VolumeMount(device_path, volume_path, mount_option):
    '''
    mount -o usrjquota=aquota.user,grpjquota=aquota.group,jqfmt=vfsv0,nomblk_io_submit,synoacl,relatime /dev/drbd-md2 /volume1
    '''

    # sanity check
    if not device_path or not volume_path or not mount_option:
        log("invalid volume mount argument: {0} {1} {2}".format(device_path, volume_path, mount_option))
        return False

    # mount volume
    cmd_arg = [BIN_MOUNT, '-o', mount_option, device_path, volume_path]
    return CmdExecutor(cmd_arg)


def Ext4QuotaOn(volume_path):
    '''
    /sbin/quotacheck -c -g -u -F vfsv0 szVolPath
    '''

    if not os.path.exists(volume_path+"/aquota.user") and not os.path.exists(volume_path+"/aquota.group"):
        return True

    cmd = ["/sbin/quotacheck", "-c", "-g", "-u", "-F", "vfsv0", volume_path]
    if not CmdExecutor(cmd):
        log("fail to execute /sbin/quotacheck {}".format(volume_path))
        return False

    return True


def BtrfsQuotaOn(volume_path):
    '''
    /sbin/btrfs quota enable szVolPath
    /sbin/btrfs usrquota enable szVolPath
    '''

    cmd = ["/sbin/btrfs", "quota", "enable", volume_path]
    if not CmdExecutor(cmd):
        log("fail to execute /sbin/btrfs quota enable {}".format(volume_path))
        return False

    cmd = ["/sbin/btrfs", "usrquota", "enable", volume_path]
    if not CmdExecutor(cmd):
        log("fail to execute /sbin/btrfs usrquota enable {}".format(volume_path))
        return False

    return True


def QuotaOn(volume_path, fs_type):
    if "ext4" == fs_type:
        return Ext4QuotaOn(volume_path)
    elif "btrfs" == fs_type:
        return BtrfsQuotaOn(volume_path)


def VolumeUnmount(volume_path):
    '''
    umount /volume1
    '''

    if not volume_path:
        log("invalid volume path".format(volume_path))
        return False

    cmd_arg = [BIN_UMOUNT, "-R", "-A", "-k", volume_path]
    return CmdExecutor(cmd_arg)


def SynoVolumeMount(device_path, volume_path):
    '''
    VolumeMount(const char *szDevPath, const char *szVolPath)
    '''

    if not device_path or not volume_path:
        log("invalid argument")
        return False

    libsynostoragecore = ctypes.cdll.LoadLibrary(LIB_SYNO_STORAGECORE)
    if not libsynostoragecore:
        log("fail to load library: {}".format(LIB_SYNO_STORAGECORE))
        return False

    libsynostoragecore.VolumeMount.restype = ctypes.c_int

    try:
        ret = libsynostoragecore.VolumeMount(device_path, volume_path)
        if 0 > ret:
            log("fail to invoke SynoVolumeMount({0}, {1}): {2}".format(device_path, volume_path, ret))
            return False
    except Exception:
        log("fail to invoke SynoVolumeMount")
        return False

    return True


def SynoVolumeUnmount(volume_path):
    '''
    VolumeUnmount(const char *szVolPath)
    '''

    if not volume_path:
        log("invalid argument")
        return False

    libsynostoragecore = ctypes.cdll.LoadLibrary(LIB_SYNO_STORAGECORE)
    if not libsynostoragecore:
        log("fail to load library: {}".format(LIB_SYNO_STORAGECORE))
        return False

    libsynostoragecore.VolumeUnmount.restype = ctypes.c_int

    try:
        ret = libsynostoragecore.VolumeUnmount(volume_path)
        if 0 > ret:
            log("fail to invoke SynoVolumeUnmount({0}): {1}".format(volume_path, ret))
            return False
    except Exception:
        log("fail to invoke SynoVolumeUnmount")
        return False

    return True


def SynoVolumeUmountWithRetry(volume_path, retry_count):
    for i in range(1, retry_count):
        if SynoVolumeUnmount(volume_path):
            return True

        time.sleep(1)

    return False


def SYNOLocalAccountUserPrepareForPkg(user_name, primary_group, home_path):
    '''
    int SYNOLocalAccountUserPrepareForPkg(const char *szUserName, const char *szPrimaryGroup, const char *szHomePath)
    '''

    if not user_name or not primary_group or not home_path:
        log("invalid argument")
        return False

    libsynolocalaccount = ctypes.cdll.LoadLibrary(LIB_SYNO_LOCALACCOUNT)
    if not libsynolocalaccount:
        log("fail to load library: {}".format(LIB_SYNO_LOCALACCOUNT))
        return False

    libsynolocalaccount.SYNOLocalAccountUserPrepareForPkg.restype = ctypes.c_int

    try:
        ret = libsynolocalaccount.SYNOLocalAccountUserPrepareForPkg(user_name, primary_group, home_path)
        if 0 > ret:
            log("fail to invoke SYNOLocalAccountUserPrepareForPkg({0}, {1}, {2}): {3}".format(
                user_name, primary_group, home_path, ret))
            return False
    except Exception:
        log("fail to invoke SYNOLocalAccountUserPrepareForPkg")
        return False

    return True

### user relative api ###


def execWebAPI(api, version, method, **kwargs):
    cmd = ["/usr/syno/bin/synowebapi", "--exec"]
    cmd.append("api=" + api)
    cmd.append("version=" + str(version))
    cmd.append("method=" + method)
    for key, val in kwargs.items():
        cmd.append('{0}={1}'.format(key, json.dumps(val)))

    try:
        with open('/dev/null', 'w') as null_fp:
            raw_resp = subprocess.check_output(cmd, stderr=null_fp)
        return json.loads(raw_resp)
    except Exception:
        return False


def IsUserExisted(user_name):
    '''
    /usr/syno/bin/synowebapi --exec api=SYNO.Core.User method=get name="MigrationAssistant"
    '''

    # sanity check
    if not user_name:
        log("invalid argument {0}".format(user_name))
        return False

    # run webapi
    j_data = execWebAPI('SYNO.Core.User', 1, 'get', name=user_name)
    if not j_data:
        return False

    # check response
    try:
        if j_data['success']:
            return True
    except Exception:
        pass

    return False


def UserCreate(user_name, pswd, desc, retry_count=3):
    '''
    /usr/syno/bin/synowebapi --exec api=SYNO.Core.User method=create name="MigrationAssistant" password="q"
    '''

    # sanity check
    if not user_name or not pswd or not desc:
        log("invalid argument {0}, {1}, {2}".format(user_name, pswd, desc))
        return False

    # run webapi
    kwargs = {
        'name': user_name,
        'password': pswd,
        'description': desc
    }

    success = False
    for i in range(1, retry_count):
        j_data = execWebAPI('SYNO.Core.User', 1, 'create', **kwargs)
        if not j_data:
            log("fail to create user")
            break
        elif not j_data['success']:
            log("fail to create user, error:{})".format(j_data['error']['code']))
            # retry if error code is WEBAPI_CORE_USER_ERR_EXCLUDE_LOCK
            if 3112 != j_data['error']['code']:
                break
        else:
            success = True
            break
        time.sleep(1)

    if not success:
        return False

    # use `synouser -setpw` to erase the password-must-change flag
    # it is some kind of hacks but it works very well and very simply.
    return CmdExecutor([BIN_SYNO_USER, '--setpw', user_name, pswd])


def UserDelete(user_name, retry_count=3):
    '''
    /usr/syno/bin/synowebapi --exec api=SYNO.Core.User method=delete name="MigrationAssistant"
    '''

    if not user_name:
        log("invalid argument {0}".format(user_name))
        return False

    # run webapi
    for i in range(1, retry_count):
        j_data = execWebAPI('SYNO.Core.User', 1, 'delete', name=[user_name])
        if not j_data:
            log("fail to delete user")
            return False
        elif not j_data['success']:
            log("fail to delete user, error:{})".format(j_data['error']['code']))
            # retry if error code is WEBAPI_CORE_USER_ERR_EXCLUDE_LOCK
            if 3112 != j_data['error']['code']:
                return False
        else:
            return True
        time.sleep(1)

    return False


def UserGroupJoinByWebAPI(user_name, group_name):
    '''
    /usr/syno/bin/synowebapi --exec api=SYNO.Core.User.Group method=join name="MigrationAssistant" join_group="administrators"
    '''

    if not user_name or not group_name:
        log("invalid argument {0}, {1}".format(user_name, group_name))
        return False

    # run webapi
    j_data = execWebAPI('SYNO.Core.User.Group', 1, 'join', name=user_name, join_group=group_name)
    if not j_data:
        return False

    return j_data['success']


def GroupMemberList(group_name):
    '''
    Parsing output of synogroup:
    /usr/syno/sbin/synogroup --get administrators
    '''

    user_list = []

    get_cmd_arg = [BIN_SYNO_GROUP, "--get", group_name]
    group_info = subprocess.Popen(get_cmd_arg, stdout=subprocess.PIPE)

    # read out first 4 lines
    line = group_info.stdout.readline() #Group Name: [administrators]
    line = group_info.stdout.readline() #Group Type: [AUTH_LOCAL]
    line = group_info.stdout.readline() #Group ID:   [101]
    line = group_info.stdout.readline() #Group Members:

    # get members
    while True:

        line = group_info.stdout.readline()  # 0:[admin]
        if not line:
            break

        if -1 != line.find(":["):
            idx = line.index(":[")

        user_name = line[idx+2:-2]
        user_list.append(user_name)

    return user_list


def UserGroupJoinByBinary(user_name, group_name):
    '''
    /usr/syno/sbin/synogroup --member administrators admin MigrationAssistant

    '''

    user_list = GroupMemberList(group_name)

    add_cmd_arg = [BIN_SYNO_GROUP, "--member", group_name, user_name] + user_list

    return CmdExecutor(add_cmd_arg)


def UserGroupJoin(user_name, group_name):

    # COMPATIBIITY: DSM6.0 don't have group join webapi
    buildnumber = int(SynoVersionGet())

    if buildnumber > DSM_61_START_VERSION:
        return UserGroupJoinByWebAPI(user_name, group_name)
    else:
        return UserGroupJoinByBinary(user_name, group_name)


def UserGroupLeave(user_name, group_name):
    '''
    /usr/syno/bin/synowebapi --exec api=SYNO.Core.User.Group method=join name=MigrationAssistant leave_group=administrators
    '''

    if not user_name or not group_name:
        log("invalid argument {0}, {1}".format(user_name, group_name))
        return False

    # run webapi
    j_data = execWebAPI('SYNO.Core.User.Group', 1, 'join', name=user_name, leave_group=group_name)
    if not j_data:
        return False

    return j_data['success']


def UserHomeCheck(user_name):
    for line in open("/etc/passwd"):
        if "/var/services/homes/{}/".format(user_name) in line:
            return True
    return False


def UserHomeDisable(user_name):
    '''
    sed -ie 's/\\/var\\/services\\/homes\\/MigrationAssistant/\\/dev\\/null/g' /etc/passwd
    '''

    cmd = [BIN_CP, "/etc/passwd", "/etc/passwd.tmp"]
    if not CmdExecutor(cmd):
        return False

    cmd = [BIN_SED, "-ie", "s/\\/var\\/services\\/homes\\/{}/\\/dev\\/null/g".format(user_name), "/etc/passwd.tmp"]
    if not CmdExecutor(cmd):
        return False

    cmd = [BIN_CP, "/etc/passwd.tmp", "/etc/passwd"]
    if not CmdExecutor(cmd):
        return False

    return True


def FileSync():
    import ctypes
    libc = ctypes.CDLL("libc.so.6")
    libc.sync()


FILE_SPACE_PROGRESS = "/tmp/space/space_building_1_0"
FILE_SPACE_PROGRESS_DIR = "/tmp/space/"
FILE_SPACE_PROGRESS_PREFIX = "/tmp/space/space_building_1_"


def SpaceProgressExisted(order):
    progress_file = FILE_SPACE_PROGRESS_PREFIX + str(order)
    return os.path.exists(progress_file)


def SpaceProgressReg(mount_point, order):
    if not mount_point:
        return False

    progress_file = FILE_SPACE_PROGRESS_PREFIX + str(order)

    with open(progress_file, 'w+') as file:
        file.write('pid=1\n')
        file.write('path={}\n'.format(mount_point))
        file.write('space_type=1\n')
        file.write('device_type=0\n')
        file.write('step=14\n')
        file.write('build_task=28\n')
        file.write('building_devices=\n')
        file.write('raid_level=0\n')
        file.write('is_add_spare=0\n')
        file.write('pool_path=\n')
        file.write('reuse_space_id=\n')
        file.write('target_space_path=\n')
        file.write('disks=\n')

    if not os.path.isfile(progress_file):
        return False

    return True


def SpaceProgressUnreg(order):

    progress_file = FILE_SPACE_PROGRESS_PREFIX + str(order)

    if os.path.exists(progress_file):
        os.unlink(progress_file)

    return True


def SpaceProgressUnregAll():
    for file in os.listdir(FILE_SPACE_PROGRESS_DIR):
        path = FILE_SPACE_PROGRESS_DIR + file
        if path.startswith(FILE_SPACE_PROGRESS_PREFIX):
            os.unlink(path)

    return True


def SpaceStatusGet():
    try:
        space = {}
        cmd = "/usr/syno/bin/synowebapi --exec api=SYNO.Storage.CGI.Storage method=load_info"
        output, err = subprocess.Popen(cmd.split(), stdout=subprocess.PIPE).communicate()
        storage = json.loads(output.rstrip())

        for vol in storage["data"]["volumes"]:
            space[vol["vol_path"]] = vol["status"]
        for lun in storage["data"]["iscsiLuns"]:
            if "block" == lun["iscsi_lun"]["device_type"]:
                space[lun["iscsi_lun"]["name"]] = lun["status"]

        return space
    except:
        return {}


def IsVolumeOnDRBD(volume_path):
    volume_info = SynoMountInfoGetByVolume(volume_path)

    if volume_info and "drbd" in volume_info['device_path']:
        return True
    return False


def VolumeList():
    try:
        cmd = "/usr/syno/bin/synowebapi --exec api=SYNO.Storage.CGI.Storage method=load_info"
        output, err = subprocess.Popen(cmd.split(), stdout=subprocess.PIPE).communicate()
        storage = json.loads(output.rstrip())

        return storage["data"]["volumes"]
    except:
        return []

# vim: set ts=4 sts=4 sw=4 et:
