#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright (C) 2000-2019 Synology Inc. All rights reserved.
#
# Created by:   jamiec
# Created on:   2019-01-03

import sys
import sdk
import drbd
import time
import subprocess
import json

# return values of following functions:
#  - LoadSyncLayer
#  - UnloadSyncLayer
#  - LoadAllSyncLayer
#  - ReloadAllSyncLayer
#  - UnloadAllSyncLayer
RET_SUCCESS = 0
RET_UNKNOWN_ERROR = 1
RET_INVALID_ARGS = 2
RET_STOP_SERVICE_FAIL = 3
RET_START_SERVICE_FAIL = 4
RET_MOUNT_VOLUME_FAIL = 5
RET_UMOUNT_VOLUME_FAIL = 6
RET_SETUP_DRBD_FAIL = 7
RET_POOL_CRASHED = 8

def LoadSyncLayer(space_refer, port, service_control=True):

    # get volume info by reading config
    vol_info = drbd.VolumeInfoGetBySpaceRefer(space_refer)
    if not vol_info:
        sdk.log("fail to get vol info")
        return RET_UNKNOWN_ERROR

    org_device_path = str(vol_info['dev_path'])
    org_mount_option = str(vol_info['mnt_opt'])
    is_volume_type = ("volume" == str(vol_info['type']))
    volume_path = space_refer

    # Do feasibility check
    if not sdk.SynoFeasibilityCheck():
        sdk.log("fail to do feasibility check")
        return RET_UNKNOWN_ERROR

    # 1. Stop all services
    if service_control and not sdk.SynoServiceStopAll():
        sdk.log("fail to stop services")
        return RET_STOP_SERVICE_FAIL

    # TODO: if volume is not mounted, consider to skip following operation

    # 2. Umount volume
    sdk.log("1. Umount volume: {}".format(volume_path))
    if is_volume_type and not sdk.SynoVolumeUmountWithRetry(volume_path, 5):
        sdk.log("fail to umount volume")

        # recover status
        if service_control:
            sdk.SynoServiceStartAll()

        return RET_UMOUNT_VOLUME_FAIL

    # 3. Create drbd device
    sdk.log("2. Create drbd device for {}".format(volume_path))
    drbd_device_path = drbd.ResourceCreate(volume_path, org_device_path, port)
    if not drbd_device_path:
        sdk.log("fail to create drbd resource")

        # recover status
        if is_volume_type:
            sdk.SynoVolumeMount(org_device_path, volume_path)
        if service_control:
            sdk.SynoServiceStartAll()

        return RET_SETUP_DRBD_FAIL

    # 4. Mount volume on top of drbd device
    sdk.log("3. Mount volume: {} on {} by {}".format(volume_path, drbd_device_path, org_mount_option))
    if is_volume_type:
        if not sdk.SynoVolumeMount(drbd_device_path, volume_path):
            sdk.log("fail to mount volume")
            return RET_MOUNT_VOLUME_FAIL

    # 5. Start all services
    if service_control and not sdk.SynoServiceStartAll():
        sdk.log("fail to start services")
        return RET_START_SERVICE_FAIL

    return RET_SUCCESS


def UnloadSyncLayer(space_refer, service_control=True):

    # get volume info by reading config
    vol_info = drbd.VolumeInfoGetBySpaceRefer(space_refer)
    if not vol_info:
        sdk.log("fail to get vol info")
        return RET_UNKNOWN_ERROR

    org_device_path = str(vol_info['dev_path'])
    is_volume_type = ("volume" == str(vol_info['type']))
    volume_path = space_refer

    # skip if volume is not mounted on DRBD
    if is_volume_type and not sdk.IsVolumeOnDRBD(space_refer):
        return RET_SUCCESS

    # 1. Stop all services
    if service_control and not sdk.SynoServiceStopAll():
        sdk.log("fail to stop services")
        return RET_STOP_SERVICE_FAIL

    # 2. Umount volume
    sdk.log("1. Umount volume: {}".format(volume_path))
    if is_volume_type and not sdk.SynoVolumeUmountWithRetry(volume_path, 5):
        sdk.log("fail to umount volume")

        # recover status
        if service_control:
            sdk.SynoServiceStartAll()

        return RET_UMOUNT_VOLUME_FAIL

    # 3. Stop drbd
    sdk.log("2. Remove drbd device for {}".format(volume_path))
    if not drbd.ResourceRemove(volume_path, org_device_path):
        sdk.log("fail to remove drbd resource, but still carry on")

    # 4. Mount volume on original device path
    sdk.log("3. Mount volume: {} on {} ".format(volume_path, org_device_path))
    if is_volume_type and not sdk.SynoVolumeMount(org_device_path, volume_path):
        sdk.log("fail to mount volume")
        return RET_MOUNT_VOLUME_FAIL

    # 5. Start all services
    if service_control and not sdk.SynoServiceStartAll():
        sdk.log("fail to start services")
        return RET_START_SERVICE_FAIL

    return RET_SUCCESS


def LoadAllSyncLayer(dev_list):

    # stop all services
    sdk.log("start to stop services")
    if not sdk.SynoServiceStopAll():
        sdk.log("fail to stop services")
        return RET_STOP_SERVICE_FAIL

    # load all drbd devices
    for dev in dev_list:

        order = int(dev["port"])-7400
        if sdk.SpaceProgressExisted(order):
            sdk.log("skip load sync layer {0}".format(dev["refer"]))
            continue

        if not sdk.SpaceProgressReg(dev["refer"], int(dev["port"])-7400):
            sdk.log("fail to register space progress")
            sdk.SynoServiceStartAll()
            return RET_UNKNOWN_ERROR

        ret = LoadSyncLayer(dev["refer"], dev["port"], False)
        if RET_SUCCESS != ret:
            sdk.log("fail to load sync layer {0} {1}".format(dev["refer"], dev["port"]))
            sdk.SynoServiceStartAll()
            return ret

    # start all services
    sdk.log("start to start services")
    if not sdk.SynoServiceStartAll():
        sdk.log("fail to start services")
        return RET_START_SERVICE_FAIL

    return RET_SUCCESS


def UnloadAllSyncLayer(dev_list):

    # stop all services
    sdk.log("start to stop services")
    if not sdk.SynoServiceStopAll():
        sdk.log("fail to stop services")
        return RET_STOP_SERVICE_FAIL

    # get current space status
    curSpaceStatus = sdk.SpaceStatusGet()

    # unload all drbd devices
    for dev in dev_list:
        # skip if the space is crashed or deleted
        if "crashed" == curSpaceStatus.get(dev["refer"], "crashed"):
            continue

        ret = UnloadSyncLayer(dev["refer"], False)
        if RET_SUCCESS != ret:
            sdk.log("fail to unload sync layer {0}".format(dev["refer"]))
            sdk.SynoServiceStartAll()
            sdk.SpaceProgressUnregAll()
            return ret

    # unregister all space progress
    if not sdk.SpaceProgressUnregAll():
        sdk.log("fail to unregister all space progress")
        sdk.SynoServiceStartAll()
        return RET_UNKNOWN_ERROR

    # start all services
    sdk.log("start to start services")
    if not sdk.SynoServiceStartAll():
        sdk.log("fail to start services")
        return RET_START_SERVICE_FAIL

    return RET_SUCCESS


def DisableAllServices():
    if not sdk.SynoServiceDisableAll():
        sdk.log("fail to disable all service")
        return RET_STOP_SERVICE_FAIL

    return RET_SUCCESS


def ReloadAllSyncLayer(dev_list):

    # stop all services
    sdk.log("start to stop services")
    if not sdk.SynoServiceStopAll():
        sdk.log("fail to stop services")
        return RET_STOP_SERVICE_FAIL

    # unload all drbd devices
    for dev in dev_list:

        ret = UnloadSyncLayer(dev["refer"], False)
        if RET_SUCCESS != ret:
            sdk.log("fail to unload sync layer {0}".format(dev["refer"]))
            sdk.SynoServiceStartAll()
            return ret

        if not sdk.SpaceProgressUnreg(int(dev["port"])-7400):
            sdk.log("fail to unregister space progress".format(dev["refer"]))
            sdk.SynoServiceStartAll()
            return RET_UNKNOWN_ERROR

    # load all drbd devices
    for dev in dev_list:

        if not sdk.SpaceProgressReg(dev["refer"], int(dev["port"])-7400):
            sdk.log("fail to register space progress")
            sdk.SynoServiceStartAll()
            return RET_UNKNOWN_ERROR

        ret = LoadSyncLayer(dev["refer"], dev["port"], False)
        if RET_SUCCESS != ret:
            sdk.log("fail to load sync layer {0} {1}".format(dev["refer"], dev["port"]))
            sdk.SynoServiceStartAll()
            return ret

    # start all services
    sdk.log("start to start services")
    if not sdk.SynoServiceStartAll():
        sdk.log("fail to start services")
        return RET_START_SERVICE_FAIL

    return RET_SUCCESS


def CheckAllSyncLayer(dev_list):

    for dev in dev_list:
        # check all space progress files are existed
        order = int(dev["port"])-7400
        if not sdk.SpaceProgressExisted(order):
            return False

    return True


def get_dev_list_by_arg(dev_arg):
    dev_list = []

    # parsing arguments into list of dict
    for entry in dev_arg.split(","):
        dev = entry.split("=")
        #sdk.log("args: {0} {1}".format(dev[0], dev[1]))
        dev_list.append({"refer":dev[0], "port":dev[1]})

    return dev_list


def record_dev_arg(dev_arg):
    # record dev_arg to /usr/syno/etc/synomigration/dev_arg
    try:
        with open(sdk.SZF_DEV_ARG, "w") as file:
            file.write(dev_arg)
        return True
    except:
        return False


def load_all_drbd():
    dev_arg = sys.argv[2]
    if not dev_arg:
        exit(RET_INVALID_ARGS)

    if not record_dev_arg(dev_arg):
        sdk.log("Failed to record dev_arg [{}] to [{}]".format(dev_arg, sdk.SZF_DEV_ARG))

    dev_list = get_dev_list_by_arg(dev_arg)
    ret = LoadAllSyncLayer(dev_list)
    exit(ret)


def unload_all_drbd():
    dev_arg = sys.argv[2]
    if not dev_arg:
        exit(RET_INVALID_ARGS)

    dev_list = get_dev_list_by_arg(dev_arg)
    ret = UnloadAllSyncLayer(dev_list)
    exit(ret)


def disable_all_services():
    ret = DisableAllServices()
    exit(ret)


def reload_all_drbd():
    dev_arg = sys.argv[2]
    if not dev_arg:
        exit(RET_INVALID_ARGS)

    dev_list = get_dev_list_by_arg(dev_arg)
    ret = ReloadAllSyncLayer(dev_list)
    exit(ret)


def check_all_drbd():
    dev_arg = sys.argv[2]
    if not dev_arg:
        exit(RET_INVALID_ARGS)

    dev_list = get_dev_list_by_arg(dev_arg)
    if not CheckAllSyncLayer(dev_list):
        exit(RET_UNKNOWN_ERROR)

    exit(RET_SUCCESS)


def check_pool():
    cmd = '/usr/syno/bin/synowebapi --exec api=SYNO.Core.Storage.Pool method=list offset=0 limit=-1'
    ret = RET_SUCCESS
    try:
        output, err = subprocess.Popen(cmd.split(), stdout=subprocess.PIPE).communicate()
        pools = json.loads(output.rstrip())
        for pool in pools["data"]["pools"]:
            if "crashed" == pool.get("status", ""):
                sdk.log("the storage pool [{}] is crashed".format(pool.get("id", "")))
                ret = RET_POOL_CRASHED
                break
    except:
        ret = RET_UNKNOWN_ERROR

    exit(ret)


if __name__ == '__main__':
    subcommand = {
        'load_all': load_all_drbd,
        'reload_all': reload_all_drbd,
        'unload_all': unload_all_drbd,
        'disable_all_services': disable_all_services,
        'check': check_all_drbd,
        'check_pool': check_pool
    }
    if len(sys.argv) > 1 and sys.argv[1] in subcommand:
        subcommand[sys.argv[1]]()
    else:
        sdk.log("invalid argument")
        exit(RET_INVALID_ARGS)

# vim:set shiftwidth=4 tabstop=4 expandtab:
